/*
 * Copyright (c) 2016, Freescale Semiconductor, Inc.
 * Copyright 2016-2017 NXP
 * All rights reserved.
 * 
 * SPDX-License-Identifier: BSD-3-Clause
 */

#include "TSI_key.h"
#include "fsl_tsi_v5.h"

/*******************************************************************************
 * Definitions
 ******************************************************************************/
/*******************************************************************************
 * Variables
 ******************************************************************************/
uint8_t g_keyId = 0U;

/*******************************************************************************
 * Code
 ******************************************************************************/

uint8_t TSI_keyDetect(uint8_t *current_key_id)
{
    uint8_t TSI_tx_channel, TSI_rx_channel;
    uint8_t tx_id, rx_id;

    uint16_t TSI_sampleResult = 0U;

    uint8_t key_state_temp;
    uint16_t key_release_debounce_temp;
    uint16_t key_baseline_temp;
    uint8_t key_touch_debounce_temp;

    uint16_t key_touch_threshold_temp;
    uint16_t key_idle_high;
    uint16_t key_idle_low;

    uint8_t delay_i = 0U;
    uint8_t delay_j = 0U;

    key_event_t key_event = kKey_Event_Idle;

    key_tsi_mapping_t *key;

    key = &key_TSI[g_keyId];

    TSI_tx_channel = key->TSI_channel_tx;
    TSI_rx_channel = key->TSI_channel_rx;

    if ((TSI_tx_channel == 0xFFU) || (TSI_rx_channel == 0xFFU))
    {
        /* Reach the end of key_TSI[] array. Re-start the TSI scan from the first key*/
        g_keyId = 0U;

        key = &key_TSI[g_keyId];

        TSI_tx_channel = key->TSI_channel_tx;
        TSI_rx_channel = key->TSI_channel_rx;
    }

    if ((TSI_tx_channel >= kTSI_Chnl_MAX_ID) | (TSI_rx_channel >= kTSI_Chnl_MAX_ID))
    {
        /*Set the default value if the tsi channel is illegal value.*/
        *current_key_id = 0xFFU;
        return 0U;
    }

    /*Write '1' to SWTS will start a scan on the channel specified by TSICH*/
    tx_id = TSI_tx_channel - kTSI_Chnl_0;
    rx_id = TSI_rx_channel - kTSI_Chnl_6;

    TSI_SetMutualCapTxChannel(TSI, (tsi_mutual_tx_channel_t)tx_id);
    TSI_SetMutualCapRxChannel(TSI, (tsi_mutual_rx_channel_t)rx_id);
    TSI_StartSoftwareTrigger(TSI);

    while (!(TSI_GetStatusFlags(TSI) & kTSI_EndOfScanFlag))
    {
        /*Start time-out counter*/
        delay_j++;
        if (delay_j >= 100U)
        {
            delay_j = 0U;
            delay_i++;
        }

        if (delay_i >= 100U)
        {
            /* If TSI scan do NOT ends after some time, force this round of TSI scan to an end */
            /* TSI_reinit();*/
            return 0xFFU;
        }
    }

    TSI_sampleResult = TSI_GetCounter(TSI);
    TSI_ClearStatusFlags(TSI, kTSI_EndOfScanFlag);

    if (TSI_sampleResult == 0xFFU)
    {
        /*There's an error occurred during the current tsi channel*/
        TSI_sampleResult = 0U;
    }

    key_state_temp = key->key_state;
    key_touch_debounce_temp = key->key_touch_debounce;
    key_release_debounce_temp = key->key_release_debounce;
    key_baseline_temp = key->key_baseline;
    key_touch_threshold_temp = key_baseline_temp + key->key_touch_delta;
    key_idle_high = key_baseline_temp + key->key_baseline_delta;
    key_idle_low = key_baseline_temp - key->key_baseline_delta;

    if (key_state_temp == kKey_State_Idle)
    {
        if (TSI_sampleResult > key_touch_threshold_temp)
        {
            /*There's a finger touch*/
            key_touch_debounce_temp++;
            key_release_debounce_temp = 0U;

            if (key_touch_debounce_temp > MAX_TOUCH_DEBOUNCE)
            {
                key_touch_debounce_temp = 0U;
                key_release_debounce_temp = 0U;
                key_event = kKey_Event_Touch; /*Indicates the TSI channel is touched*/
                key_state_temp = kKey_State_Touched;
            }
        }
        else if (TSI_sampleResult > key_idle_high)
        {
            key_touch_debounce_temp = 0U;
            key_release_debounce_temp++;

            if (key_release_debounce_temp > 10U)
            {
                key_release_debounce_temp = 0U;
                key_baseline_temp = key_baseline_temp + 1U;
            }
        }
        else if (TSI_sampleResult < key_idle_low)
        {
            key_touch_debounce_temp = 0U;
            key_release_debounce_temp++;

            if (key_release_debounce_temp > 10U)
            {
                key_release_debounce_temp = 0U;
                key_baseline_temp = key_baseline_temp - 1U;
            }
        }
        else
        {
            key_touch_debounce_temp = 0U;
            key_release_debounce_temp = 0U;
        }
    }
    else if (key_state_temp == kKey_State_Touched)
    {
        key_event = kKey_Event_Touched; /* Indicates the TSI channel is touched*/

        if (TSI_sampleResult > key_touch_threshold_temp)
        {
            /*Indicates the TSI channel is Stick Touched*/
            key_release_debounce_temp = 0U;
        }
        else
        {
            /*Condition: finger touch release*/
            key_touch_debounce_temp = 0U;
            key_release_debounce_temp++;

            if (key_release_debounce_temp > MAX_TOUCH_DEBOUNCE)
            {
                key_release_debounce_temp = 0U;
                key_event = kKey_Event_Release;
                key_state_temp = kKey_State_Released;
            }
        }
    }
    else if (key_state_temp == kKey_State_Released)
    {
        /*Restore kKey_State_Idle automatically*/
        key_event = kKey_Event_Released;

        key_release_debounce_temp++;

        if (key_release_debounce_temp > 10U)
        {
            /*Delay to assure this key is sent*/
            key_release_debounce_temp = 0U;
            key_event = kKey_Event_Idle;
            key_state_temp = kKey_State_Idle;
        }
    }
    else
    {
        /*Key state can NOT go here*/
        key_event = kKey_Event_Idle;
        key_state_temp = kKey_State_Idle;
    }

    key->key_state = key_state_temp;
    key->key_current_value = TSI_sampleResult;
    key->key_baseline = key_baseline_temp;
    key->key_touch_threshold = key_touch_threshold_temp;
    key->key_touch_debounce = key_touch_debounce_temp;
    key->key_release_debounce = key_release_debounce_temp;

    /*Sample the next key*/
    *current_key_id = g_keyId;
    g_keyId++;

    return key_event;
}

/* Calibrate TSI idle value, save as TSI baseline */
void TSI_keyCalibrate(void)
{
    uint8_t delay_i, delay_j;

    uint8_t tx_id, rx_id;
    uint8_t key_id = 0U;

    /* Initial baseline for each mutual touch key*/
    while ((key_TSI[key_id].key_touch_delta != 0xFFU) && (key_TSI[key_id].key_baseline_delta != 0xFFU))
    {
        /* Initiate key_TSI structure*/
        key_TSI[key_id].key_state = 0U;
        key_TSI[key_id].key_touch_debounce = 0U;
        key_TSI[key_id].key_release_debounce = 0U;
        key_TSI[key_id].key_baseline = 0U;
        key_TSI[key_id].key_current_value = 0U;

        delay_i = 0U;
        delay_j = 0U;

        tx_id = key_TSI[key_id].TSI_channel_tx - kTSI_Chnl_0;
        rx_id = key_TSI[key_id].TSI_channel_rx - kTSI_Chnl_6;

        TSI_SetMutualCapTxChannel(TSI, (tsi_mutual_tx_channel_t)tx_id);
        TSI_SetMutualCapRxChannel(TSI, (tsi_mutual_rx_channel_t)rx_id);
        TSI_StartSoftwareTrigger(TSI);

        while (!(TSI_GetStatusFlags(TSI) & kTSI_EndOfScanFlag))
        {
            /* Start time-out counter*/
            delay_j++;
            if (delay_j >= 100U)
            {
                delay_j = 0U;
                delay_i++;
            }

            if (delay_i >= 100U)
            {
                /* If TSI scan do NOT ends after some time, force this round of TSI scan to an end */
                /*TSI_reinit();*/
                return;
            }
        }

        key_TSI[key_id].key_baseline = TSI_GetCounter(TSI);
        TSI_ClearStatusFlags(TSI, kTSI_EndOfScanFlag);

        key_id++;
    }
}
